;*---------------------------------------------------------------------*/
;*    Copyright (c) 1996 by Manuel Serrano. All rights reserved.       */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \    /  '                              */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome Send them to                                           */
;*        <Manuel.Serrano@inria.fr>                                    */
;*        Manuel Serrano                                               */
;*        INRIA -- Rocquencourt                                        */
;*        Domaine de Voluceau, BP 105                                  */
;*        78153 Le Chesnay Cedex                                       */
;*        France                                                       */
;*---------------------------------------------------------------------*/


;*---------------------------------------------------------------------*/
;*    serrano/prgm/project/bigloo/examples/Depend/depend.scm ...       */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Wed Mar 17 10:49:15 1993                          */
;*    Last change :  Wed Jan 18 13:33:53 1995 (serrano)                */
;*                                                                     */
;*    On genere des dependances (d'apres les includes).                */
;*---------------------------------------------------------------------*/
 
;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module afile (main main))

;*---------------------------------------------------------------------*/
;*    main ...                                                         */
;*---------------------------------------------------------------------*/
(define (main argv)
   (if (or (null?    (cdr argv))
	   (string=? (cadr argv) "-help"))
       (usage)
       (let loop ((files        (cdr argv))
		  (files-list  '())
		  (output-file '()))
	  (cond
	     ((null? files)
	      (output files-list output-file))
	     ((string=? (car files) "-o")
	      (if (null? (cdr files))
		  (usage)
		  (loop (cddr files)
			files-list
			(cadr files))))
	     (else
	      (loop (cdr files)
		    (cons (car files) files-list)
		    output-file))))))

;*---------------------------------------------------------------------*/
;*    output ...                                                       */
;*---------------------------------------------------------------------*/
(define (output files-list output-file)
   (let ((port (if (string? output-file)
		   (begin
		      (if (file-exists? output-file)
			  (rename-file output-file
				       (string-append output-file "~")))
		      (open-output-file output-file))
		   (current-output-port))))
      (let loop ((files-list files-list))
	 (if (null? files-list)
	     (newline port)
	     (let* ((file-name (car files-list))
		    (key       (gensym))
		    (includes  (find-includes key file-name #t)))
		(if (not (null? includes))
		    (begin
		       (display (string-append (remove-extansion file-name)
					       ".o")
				port)
		       (display #\: port)
		       (for-each (lambda (i)
				    (display #\space port)
				    (display i port))
				 includes)
		       (newline port)))
		(loop (cdr files-list)))))))
	  
;*---------------------------------------------------------------------*/
;*    find-includes ...                                                */
;*---------------------------------------------------------------------*/
(define (find-includes key file w/error)
   (cond
      ((not (file-exists? file))
       (fprint (current-error-port) "*** ERROR:depend:" #\Newline
	       "Can't find file -- " file)
       '())
      ((eq? (getprop (string->symbol file) 'include) key)
       '())
      (else
       ;; on marque que le fichier a ete examine
       (putprop! (string->symbol file) 'include key)
       (let ((port (open-input-file file)))
	  (if (not (input-port? port))
	      (begin
		 (fprint (current-error-port) "*** ERROR:depend:" #\Newline
			 "Can't open file -- " file)
		 '())
	      (let ((exp (read port)))
		 (match-case exp
		    ((or (module ?- . ?clauses)
			 (directives . ?clauses))
		     (let loop ((clauses  clauses)
				(includes '()))
			(if (null? clauses)
			    (begin
			       (close-input-port port)
			       includes)
			    (if (eq? (car (car clauses)) 'include)
				(loop (cdr clauses)
				      (append (cdr (car clauses))
					      (apply append
						     (map
						      (lambda (f)
							 (find-includes key
									f
									#f))
						      (cdr (car clauses))))
					      includes))
				(loop (cdr clauses)
				      includes)))))
		    (else
		     (close-input-port port)
		     (if w/error
			 (begin
			    (fprint (current-error-port) "*** ERROR:depend:"
				    #\Newline
				    "Illegal file format -- " file)
			    '())
			 '())))))))))
	     
;*---------------------------------------------------------------------*/
;*    usage ...                                                        */
;*---------------------------------------------------------------------*/
(define (usage)
   (print "usage: depend [-o output] <file1> <file2> ... <filen>")
   (exit -1))

;*---------------------------------------------------------------------*/
;*    remove-extansion ...                                             */
;*---------------------------------------------------------------------*/
(define (remove-extansion string)
   (let ((len (-fx (string-length string) 1)))
      (let loop ((e len)
                 (s len))
         (cond
            ((=fx s 0)
             (substring string 0 (+fx 1 e)))
            (else
             (if (and (eq? (string-ref string s) #\.)
                      (=fx e len))
                 (loop (-fx s 1) (- s 1))
                 (loop e (-fx s 1))))))))
